from __future__ import annotations
from typing import AsyncIterator, Generic, Optional, TypeVar

from .subscribablequeue import SubscribableQueue
from ._userdict import UserDict

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class QueueDict(UserDict[_KT, SubscribableQueue[_VT]], Generic[_KT, _VT]):
    def __missing__(self, key: _KT) -> SubscribableQueue[_VT]:
        v = self[key] = SubscribableQueue()
        return v

    def __delitem__(self, key: _KT) -> None:
        self.data.pop(key).close()


class SubscribableDict(UserDict[_KT, _VT], Generic[_KT, _VT]):
    """Dict with async generator that yields values as they change"""

    def __init__(self, *args, **kwargs):
        self._queue: QueueDict[_KT, _VT] = QueueDict()
        super().__init__(*args, **kwargs)

    def subscribe(
        self, key: _KT, last: Optional[bool] = True
    ) -> AsyncIterator[_VT]:
        """Async generator that yields values for the key as they are set

        Waits for new values and yields them as they are set. If `last` is
        true, yields immediately the current value for the key before starting
        to wait. If the key doesn't exist, waits for the first value for the
        key; KeyError won't be raised.

        """
        return self._queue[key].subscribe(last=last)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        """Set the value for the key, yielding the value in the generators"""
        super().__setitem__(key, value)
        self._queue[key].put(value)

    def __delitem__(self, key: _KT) -> None:
        """Remove the key, ending all subscriptions for the key

        The async generators returned by the method `subscribe()` for the key
        will return.

        KeyError will be raised if the key doesn't exist
        """
        super().__delitem__(key)
        self.end(key)

    def end(self, key: _KT):
        """End all subscriptions for the key without removing the key

        The async generators returned by the method `subscribe()` for the key
        will return.

        The item for the key is still accessible as d[key]

        KeyError will not be raised

        """
        try:
            del self._queue[key]
        except KeyError:
            pass

    def close(self) -> None:
        """End all subscriptions for all keys

        The all async generators returned by the method `subscribe()` for any
        key will return.

        No keys will be removed, i.e., d[key] still returns the item for the
        key.

        """
        self._queue.clear()



from __future__ import annotations

import pytest

from nextline.state import State, Running, Finished, StateObsoleteError
from nextline.types import TraceNo

from .base import BaseTestState


class TestRunning(BaseTestState):

    state_class = Running

    @pytest.fixture()
    def state(self, running: State) -> State:
        return running

    async def assert_obsolete(self, state: State):
        assert "obsolete" in repr(state)

        with pytest.raises(StateObsoleteError):
            state.run()

        with pytest.raises(StateObsoleteError):
            await state.finish()

        with pytest.raises(StateObsoleteError):
            state.reset()

        with pytest.raises(StateObsoleteError):
            state.close()

    @pytest.mark.asyncio
    async def test_finish(self, state):
        finished = await state.finish()
        assert isinstance(finished, Finished)
        await self.assert_obsolete(state)

    def test_run_called(self, state: State, mock_run):
        del state
        assert mock_run.call_count == 1
        registry, q_commands, q_done = mock_run.call_args.args
        del registry, q_commands, q_done

    def test_send_pdb_command(self, state: State, mock_run):  # type: ignore
        trace_id = TraceNo(1)
        command = "next"
        state.send_pdb_command(trace_id, command)
        _, q_commands, _ = mock_run.call_args.args
        assert (trace_id, command) == q_commands.get()



import pytest

from nextline.state import Initialized, Running, Finished, Closed
from nextline.process.run import RunArg


@pytest.fixture
def context() -> RunArg:
    y = RunArg()
    return y


@pytest.mark.asyncio
async def test_transition(context: RunArg):

    state = Initialized(context=context)
    assert isinstance(state, Initialized)

    state = state.run()
    assert isinstance(state, Running)

    state = await state.finish()
    assert isinstance(state, Finished)

    state = state.close()
    assert isinstance(state, Closed)


@pytest.fixture(autouse=True)
def monkey_patch_run(monkey_patch_run):
    yield monkey_patch_run



from __future__ import annotations

import asyncio
import queue
import multiprocessing
from typing import Any, Tuple

# from weakref import WeakKeyDictionary

import pytest
from unittest.mock import Mock

from nextline.process.run import run, RunArg
from nextline.types import RunNo
from nextline.utils.func import to_thread


def test_q_done_on_exception(q_done, monkey_patch_run):
    del monkey_patch_run
    context = RunArg()
    q_commands = Mock()
    with pytest.raises(MockError):
        run(context, q_commands, q_done)
    assert (None, None) == q_done.get()


class MockError(Exception):
    pass


@pytest.fixture
def monkey_patch_run(monkeypatch):
    y = Mock(side_effect=MockError)
    monkeypatch.setattr("nextline.process.run._run", y)
    yield y


@pytest.mark.asyncio
async def test_one(
    expected_exception,
    context: RunArg,
    q_commands,
    q_done,
    task_send_commands,
):
    del task_send_commands
    await to_thread(run, context, q_commands, q_done)
    result, exception = q_done.get()
    assert result is None
    if expected_exception:
        with pytest.raises(expected_exception):
            raise exception
    else:
        exception is None


@pytest.fixture
async def task_send_commands(q_registrar, q_commands):
    y = asyncio.create_task(to_thread(respond_prompt, q_registrar, q_commands))
    yield y
    q_registrar.put(None)
    await y


def respond_prompt(q_registrar, q_commands):
    while (m := q_registrar.get()) is not None:
        key, value, _ = m
        if key != "prompt_info":
            continue
        prompt_info = value
        if prompt_info is None:
            continue
        if not prompt_info.open:
            continue
        q_commands.put((prompt_info.trace_no, "next"))


@pytest.fixture
def context(
    statement: str,
    q_registrar: multiprocessing.Queue[Tuple[str, Any, bool]],
) -> RunArg:
    y = RunArg(
        run_no=RunNo(1),
        statement=statement,
        filename="<string>",
        queue=q_registrar,
    )
    return y


@pytest.fixture
def q_registrar():
    return multiprocessing.Queue()


@pytest.fixture
def statement(statement_params):
    return statement_params[0]


@pytest.fixture
def expected_exception(statement_params):
    return statement_params[1]


SOURCE_ONE = """
import time
time.sleep(0.001)
""".strip()

SOURCE_COMPILE_ERROR = """
def
""".strip()

SOURCE_RUNTIME_ERROR = """
a
""".strip()

CODE_OBJECT = compile(SOURCE_ONE, "<string>", "exec")


@pytest.fixture(
    params=[
        (SOURCE_ONE, None),
        (SOURCE_COMPILE_ERROR, SyntaxError),
        (SOURCE_RUNTIME_ERROR, NameError),
        (CODE_OBJECT, None),
    ]
)
def statement_params(request):
    return request.param


@pytest.fixture
def q_commands():
    y = queue.Queue()
    return y


@pytest.fixture
def q_done():
    y = queue.Queue()
    return y



from __future__ import annotations
from collections import UserDict
from typing import (
    Generic,
    Optional,
    TypeVar,
    MutableMapping,
    KeysView,
    ValuesView,
    ItemsView,
    Protocol,
    Iterable,
    overload,
)


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_T = TypeVar("_T")
_VT_co = TypeVar("_VT_co", covariant=True)


class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]:
        ...

    def __getitem__(self, __k: _KT) -> _VT_co:
        ...


class TUserDict(UserDict, MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    """Trying to make type hints on UserDict work in Python 3.8

    Copied many lines from
    https://github.com/python/typeshed/blob/master/stdlib/typing.pyi

    It should be possible to just use the typing.pyi but not clear how to do it

    """

    def __getitem__(self, key: _KT) -> _VT:
        return super().__getitem__(key)

    @overload
    def get(self, key: _KT) -> Optional[_VT]:
        ...

    @overload
    def get(self, key: _KT, default: _VT | _T) -> _VT | _T:
        ...

    def get(self, key, default=None):
        return super().get(key, default)

    def items(self) -> ItemsView[_KT, _VT]:
        return super().items()

    def keys(self) -> KeysView[_KT]:
        return super().keys()

    def values(self) -> ValuesView[_VT]:
        return super().values()

    def __contains__(self, value: object) -> bool:
        return super().__contains__(value)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        return super().__setitem__(key, value)

    def __delitem__(self, key: _KT) -> None:
        return super().__delitem__(key)

    def clear(self) -> None:
        return super().clear()

    @overload
    def pop(self, key: _KT) -> _VT:
        ...

    @overload
    def pop(self, key: _KT, default: _VT | _T) -> _VT | _T:
        ...

    def pop(self, key, default=None):
        return super().pop(key, default)

    def popitem(self) -> tuple[_KT, _VT]:
        return super().popitem()

    @overload
    def setdefault(self: TUserDict[_KT, _T | None], key: _KT) -> _T | None:
        ...

    @overload
    def setdefault(self, key: _KT, default: _VT) -> _VT:
        ...

    def setdefault(self, key, default=None):
        return super().setdefault(key, default)

    @overload
    def update(
        self, __m: SupportsKeysAndGetItem[_KT, _VT], **kwargs: _VT
    ) -> None:
        ...

    @overload
    def update(self, __m: Iterable[tuple[_KT, _VT]], **kwargs: _VT) -> None:
        ...

    @overload
    def update(self, **kwargs: _VT) -> None:
        ...

    def update(self, *args, **kwargs):
        return super().update(*args, **kwargs)



import pytest
from unittest.mock import Mock

from nextline.utils import ExcThread
from queue import Queue


@pytest.fixture(autouse=True)
def monkey_patch_mp(monkeypatch):
    class mock_context:
        Queue = Queue
        Process = ExcThread

    monkeypatch.setattr("nextline.state._mp", mock_context)
    return


@pytest.fixture
def monkey_patch_run(monkeypatch):
    def mock_run(registry, q_commands, q_done) -> None:
        del registry, q_commands
        q_done.put((None, None))

    wrap = Mock(wraps=mock_run)
    monkeypatch.setattr("nextline.state.run", wrap)
    yield wrap



from __future__ import annotations

import asyncio
import pytest

from nextline import Nextline

STATEMENT = """
import time

time.sleep(100)
""".lstrip()


@pytest.mark.asyncio
async def test_run(nextline: Nextline):
    assert nextline.state == "initialized"

    await asyncio.gather(
        control(nextline),
        run(nextline),
    )
    assert nextline.state == "closed"


async def run(nextline: Nextline):
    await asyncio.sleep(0.01)
    await nextline.run()
    exc = nextline.exception()
    assert exc
    assert exc.__traceback__  # tblib
    with pytest.raises(KeyboardInterrupt):
        nextline.result()
    await nextline.close()


async def control(nextline: Nextline):
    async for prompt_info in nextline.subscribe_prompt_info():
        if not prompt_info.open:
            continue
        nextline.send_pdb_command(prompt_info.trace_no, "next")
        if prompt_info.event == "line" and prompt_info.line_no == 3:  # sleep()
            await asyncio.sleep(0.005)
            nextline.interrupt()


@pytest.fixture
def nextline(statement):
    return Nextline(statement)


@pytest.fixture
def statement():
    return STATEMENT



from __future__ import annotations

import os
import signal
import asyncio
from queue import Empty

import multiprocessing as mp
from threading import Event
from concurrent.futures import ThreadPoolExecutor
from tblib import pickling_support  # type: ignore
from typing import Optional, Any

from .utils import SubscribableDict, to_thread
from .process.run import run, RunArg, QueueCommands, QueueDone
from .registrar import Registrar
from .types import RunNo, TraceNo
from .count import RunNoCounter

_mp = mp.get_context("spawn")  # NOTE: monkey patched in tests

pickling_support.install()

SCRIPT_FILE_NAME = "<string>"


class Machine:
    """State machine

                 .-------------.
            .--->| Initialized |---.
    reset() |    '-------------'   |
            |      |   | run()     |
            |------'   |           |
            |          v           |
            |    .-------------.   |
            |    |   Running   |   |
            |    '-------------'   |
            |          | finish()  |
            |          |           |
            |          V           |
            |    .-------------.   |  close()  .-------------.
            '----|  Finished   |-------------->|   Closed    |
                 '-------------'               '-------------'


    Parameters
    ----------
    statement : str
        A Python code as a string.

    """

    def __init__(self, statement: str, run_no_start_from=1):
        filename = SCRIPT_FILE_NAME
        self.registry = SubscribableDict[Any, Any]()
        self._run_no = RunNo(run_no_start_from - 1)
        self._run_no_count = RunNoCounter(run_no_start_from)
        queue = _mp.Queue()
        self._registrar = Registrar(self.registry, queue)

        self.context = RunArg(
            statement=statement, filename=filename, queue=queue
        )

        self._registrar.script_change(script=statement, filename=filename)

        self._lock_finish = asyncio.Condition()
        self._lock_close = asyncio.Condition()

        self._state: State = Initialized(self.context)
        self._state_changed()

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f"<{self.__class__.__name__} {self.state_name!r}>"

    def _state_changed(self) -> None:
        self._registrar.state_change(self._state)
        if self._state.name == "initialized":
            self._run_no = self._run_no_count()
            self.context["run_no"] = self._run_no
            self._registrar.state_initialized(self._run_no)
        elif self._state.name == "running":
            self._registrar.run_start(self._run_no)
        elif self._state.name == "finished":
            self._registrar.run_end(state=self._state)

    @property
    def state_name(self) -> str:
        """e.g., "initialized", "running","""
        try:
            return self._state.name
        except BaseException:
            return "unknown"

    def run(self) -> None:
        """Enter the running state"""
        self._state = self._state.run()
        self._state_changed()

    def send_pdb_command(self, trace_id: int, command: str) -> None:
        self._state.send_pdb_command(trace_id, command)

    def interrupt(self) -> None:
        self._state.interrupt()

    def terminate(self) -> None:
        self._state.terminate()

    async def finish(self) -> None:
        """Enter the finished state"""
        async with self._lock_finish:
            self._state = await self._state.finish()
            self._state_changed()

    def exception(self) -> Optional[Exception]:
        ret = self._state.exception()
        return ret

    def result(self) -> Any:
        return self._state.result()

    def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ) -> None:
        """Enter the initialized state"""
        if statement:
            self.context["statement"] = statement
            self._registrar.script_change(
                script=statement, filename=SCRIPT_FILE_NAME
            )
        if run_no_start_from is not None:
            self._run_no_count = RunNoCounter(run_no_start_from)
            # self._registrar.reset_run_no_count(run_no_start_from)
        self._state = self._state.reset()
        self._state_changed()

    async def close(self) -> None:
        """Enter the closed state"""
        async with self._lock_close:
            self._state = self._state.close()
            self._state_changed()
            await to_thread(self._registrar.close)
            await to_thread(self.registry.close)


class StateObsoleteError(Exception):
    """Operation on an obsolete state object."""

    pass


class StateMethodError(Exception):
    """Irrelevant operation on the state."""

    pass


class ObsoleteMixin:
    def assert_not_obsolete(self):
        if self.is_obsolete():
            raise StateObsoleteError(f"The state object is obsolete: {self!r}")

    def is_obsolete(self):
        return getattr(self, "_obsolete", False)

    def obsolete(self):
        self._obsolete = True


class State(ObsoleteMixin):
    """The base state class in the Nextline state machine"""

    name = "state"

    def __repr__(self):
        # e.g., "<Initialized 'initialized'>"
        items = [self.__class__.__name__, repr(self.name)]
        if self.is_obsolete():
            items.append("obsolete")
        return f'<{" ".join(items)}>'

    def run(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def finish(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def reset(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def close(self) -> State:
        self.assert_not_obsolete()
        print(self.name)
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def send_pdb_command(self, trace_id: TraceNo, command: str) -> None:
        del trace_id, command
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def interrupt(self) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def terminate(self) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def exception(self) -> Optional[Exception]:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def result(self) -> Any:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")


class Initialized(State):
    """The state "initialized", ready to run

    Parameters
    ----------
    context : dict
        An instance of Context

    """

    name = "initialized"

    def __init__(self, context: RunArg):
        self._context = context

    def run(self):
        self.assert_not_obsolete()
        running = Running(self._context)
        self.obsolete()
        return running

    def reset(self):
        self.assert_not_obsolete()
        initialized = Initialized(context=self._context)
        self.obsolete()
        return initialized

    def close(self):
        self.assert_not_obsolete()
        closed = Closed()
        self.obsolete()
        return closed


class Running(State):
    """The state "running", the script is being executed.

    Parameters
    ----------
        An instance of Context
    """

    name = "running"

    def __init__(self, context: RunArg):
        self._context = context
        self._q_commands: QueueCommands = _mp.Queue()
        self._event = Event()
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._f = self._executor.submit(self._run)
        assert self._event.wait(2.0)

    def _run(self):
        q_done: QueueDone = _mp.Queue()
        self._p = _mp.Process(
            target=run,
            args=(self._context, self._q_commands, q_done),
            daemon=True,
        )
        self._p.start()
        self._event.set()
        self._p.join()
        try:
            return q_done.get(timeout=0.1)
        except Empty:
            return None, None  # ret, exc

    async def finish(self):
        self.assert_not_obsolete()
        ret, exc = await to_thread(self._f.result)
        print(type(exc), exc)
        print(self._p)
        self._executor.shutdown()
        finished = Finished(self._context, result=ret, exception=exc)
        self.obsolete()
        return finished

    def send_pdb_command(self, trace_id: TraceNo, command: str) -> None:
        self._q_commands.put((trace_id, command))

    def interrupt(self) -> None:
        if self._p.pid:
            print(self._p.pid)
            os.kill(self._p.pid, signal.SIGINT)

    def terminate(self) -> None:
        self._p.terminate()


class Finished(State):
    """The state "finished", the script execution has finished

    The thread which executed the script has been joined.

    Parameters
    ----------
    context : dict
        An instance of Context
    result : any
        The result of the script execution, always None
    exception : exception or None
        The exception of the script execution if any. Otherwise None

    """

    name = "finished"

    def __init__(self, context: RunArg, result, exception):
        self._result = result
        self._exception = exception

        self._context = context

    def exception(self):
        """Return the exception of the script execution

        Return None if no exception has been raised.

        """
        return self._exception

    def result(self):
        """Return the result of the script execution

        None in the current implementation as the build-in function
        exec() returns None.

        Re-raise the exception if an exception has been raised in the
        script execution.

        """

        if self._exception:
            raise self._exception

        return self._result

    async def finish(self):
        # This method can be called when Nextline.finish() is
        # asynchronously called multiple times.
        self.assert_not_obsolete()
        return self

    def reset(self):
        self.assert_not_obsolete()
        initialized = Initialized(context=self._context)
        self.obsolete()
        return initialized

    def close(self):
        self.assert_not_obsolete()
        closed = Closed()
        self.obsolete()
        return closed


class Closed(State):
    """The state "closed" """

    name = "closed"

    def close(self):
        self.assert_not_obsolete()
        return self



from __future__ import annotations

from asyncio import Task  # noqa F401
from multiprocessing import Queue
from threading import Thread  # noqa F401
import dataclasses
import datetime
import traceback
import json

from typing import TYPE_CHECKING, Any, Tuple
from typing import MutableMapping  # noqa F401
from typing_extensions import TypeAlias

from .types import RunNo, RunInfo
from .types import TraceInfo  # noqa F401
from .utils import ExcThread

if TYPE_CHECKING:
    from .state import State

SCRIPT_FILE_NAME = "<string>"

RunNoMap: TypeAlias = "MutableMapping[Task | Thread, int]"
TraceNoMap: TypeAlias = "MutableMapping[Task | Thread, int]"
TraceInfoMap: TypeAlias = "MutableMapping[int, TraceInfo]"


class Registrar:
    def __init__(
        self, registry: MutableMapping, queue: Queue[Tuple[str, Any, bool]]
    ):
        self._registry = registry
        self._queue = queue
        self._thread = ExcThread(target=self._relay, daemon=True)
        self._thread.start()

    def close(self):
        self._queue.put(None)
        self._thread.join()

    def _relay(self) -> None:
        while (m := self._queue.get()) is not None:
            key, value, close = m
            if close:
                try:
                    del self._registry[key]
                except KeyError:
                    pass
                continue
            self._registry[key] = value

    def script_change(self, script: str, filename: str) -> None:
        self._registry["statement"] = script
        self._registry["script_file_name"] = filename

    def state_change(self, state: State) -> None:
        self._registry["state_name"] = state.name

    def state_initialized(self, run_no: int) -> None:
        self._registry["run_no"] = run_no

    def run_start(self, run_no: RunNo) -> None:
        self._run_info = RunInfo(
            run_no=run_no,
            state="running",
            script=self._registry["statement"],
            started_at=datetime.datetime.now(),
        )
        self._registry["run_info"] = self._run_info

    def run_end(self, state: State) -> None:
        exc = state.exception()
        ret = state.result() if not exc else None
        if exc:
            fmt_exc = "".join(
                traceback.format_exception(type(exc), exc, exc.__traceback__)
            )
        else:
            ret = json.dumps(ret)
            fmt_exc = None
        self._run_info = dataclasses.replace(
            self._run_info,
            state="finished",
            result=ret,
            exception=fmt_exc,
            ended_at=datetime.datetime.now(),
        )
        # TODO: check if run_no matches
        self._registry["run_info"] = self._run_info



from collections import UserDict
from typing import TypeVar


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


try:
    UserDict[_KT, _VT]
except TypeError:
    # For Python 3.8
    from ._tuserdict import TUserDict as UserDict



import pytest

from nextline.state import State, Closed

from .base import BaseTestState


class TestClosed(BaseTestState):

    state_class = Closed

    @pytest.fixture()
    def state(self, closed: State) -> State:
        return closed

    def test_close(self, state: State):
        # The same object should be returned no matter
        # how many times called.
        assert state is state.close()
        assert state is state.close()
        assert state is state.close()
        assert "obsolete" not in repr(state)



from __future__ import annotations

from multiprocessing import Queue
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from logging import getLogger

from typing import Callable, Set, TypeVar, TypedDict, MutableMapping
from typing import Any, Tuple  # noqa F401
from typing_extensions import TypeAlias

from ..types import RunNo, TraceNo
from .trace import Trace
from .call import call_with_trace
from .callback import Callback, RegistrarProxy
from . import script

_T = TypeVar("_T")

QueueDone: TypeAlias = "Queue[Tuple[Any, BaseException | None]]"

PdbCommand: TypeAlias = str
QueueCommands: TypeAlias = "Queue[Tuple[TraceNo, PdbCommand] | None]"
PdbCiMap: TypeAlias = MutableMapping[TraceNo, Callable[[PdbCommand], Any]]


class RunArg(TypedDict, total=False):
    run_no: RunNo
    statement: str
    filename: str
    queue: Queue[Tuple[str, Any, bool]]


class Context(TypedDict):
    callback: Callback
    pdb_ci_map: PdbCiMap
    modules_to_trace: Set[str]


def run(run_arg: RunArg, q_commands: QueueCommands, q_done: QueueDone):
    try:
        _run(run_arg, q_commands, q_done)
    except BaseException:
        q_done.put((None, None))
        raise


def _run(run_arg: RunArg, q_commands: QueueCommands, q_done: QueueDone):

    run_no = run_arg["run_no"]
    statement = run_arg.get("statement")
    filename = run_arg.get("script_file_name", "<string>")
    queue = run_arg["queue"]

    try:
        code = _compile(statement, filename)
    except BaseException as e:
        q_done.put((None, e))
        return

    pdb_ci_map: PdbCiMap = {}
    modules_to_trace: Set[str] = set()

    with Callback(
        run_no=run_no,
        registrar=RegistrarProxy(queue),
        modules_to_trace=modules_to_trace,
    ) as callback:

        context = Context(
            callback=callback,
            pdb_ci_map=pdb_ci_map,
            modules_to_trace=modules_to_trace,
        )

        trace = Trace(context=context)

        func = script.compose(code)

        with relay_commands(q_commands, pdb_ci_map):
            result, exception = call_with_trace(func, trace)

    q_done.put((result, exception))


def _compile(code, filename):
    if isinstance(code, str):
        code = compile(code, filename, "exec")
    return code


@contextmanager
def relay_commands(q_commands: QueueCommands, pdb_ci_map: PdbCiMap):
    def fn():
        while m := q_commands.get():
            trace_id, command = m
            pdb_ci = pdb_ci_map[trace_id]
            pdb_ci(command)

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(try_again_on_error, fn)
        try:
            yield
        finally:
            q_commands.put(None)
            future.result()


def try_again_on_error(func: Callable[[], _T]) -> _T:
    while True:
        try:
            return func()
        # except KeyboardInterrupt:
        #     raise
        except BaseException:
            logger = getLogger(__name__)
            logger.exception("")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator, Generator, Tuple

import pytest
from unittest.mock import Mock

from nextline.state import (
    State,
    Initialized,
    Running,
    Finished,
    Closed,
    StateObsoleteError,
    StateMethodError,
)
from nextline.process.run import QueueCommands, QueueDone, RunArg
from nextline.types import TraceNo


class BaseTestState(ABC):
    """Test state classes of the state machine

    To be inherited by the test class for each state class.
    """

    state_class: Any = None

    @pytest.fixture
    def mock_run_exception(self) -> Any:
        return None

    @pytest.fixture
    def mock_run_result_exception(self, mock_run_exception) -> Tuple[Any, Any]:
        result = None
        return result, mock_run_exception

    @pytest.fixture(autouse=True)
    def mock_run(
        self,
        monkeypatch,
        mock_run_result_exception: Tuple[Any, Any],
    ):
        def run(
            run_arg: RunArg,
            q_commands: QueueCommands,
            q_done: QueueDone,
        ) -> None:
            del run_arg, q_commands
            q_done.put(mock_run_result_exception)

        wrap = Mock(wraps=run)
        monkeypatch.setattr("nextline.state.run", wrap)
        return wrap

    @pytest.fixture
    def context(self) -> RunArg:
        y = RunArg()
        return y

    @pytest.fixture()
    def initialized(
        self, context: RunArg
    ) -> Generator[Initialized, None, None]:
        y = Initialized(context=context)
        yield y
        if y.is_obsolete():
            return
        y.close()

    @pytest.fixture()
    async def running(
        self, initialized: Initialized
    ) -> AsyncGenerator[Running, None]:
        y = initialized.run()
        yield y
        if y.is_obsolete():
            return
        finished = await y.finish()
        finished.close()

    @pytest.fixture()
    async def finished(
        self, running: Running
    ) -> AsyncGenerator[Finished, None]:
        y = await running.finish()
        yield y
        if y.is_obsolete():
            return
        y.close()

    @pytest.fixture()
    def closed(self, finished: Finished) -> Closed:
        y = finished.close()
        return y

    @abstractmethod
    def state(self, _):
        """Yield an instance of the class being tested

        To be overridden as a pytest fixture.
        """
        pass

    def test_state(self, state: State):
        assert self.state_class is not None
        assert isinstance(state, self.state_class)
        assert "obsolete" not in repr(state)

    async def assert_obsolete(self, state: State):
        assert "obsolete" in repr(state)

        with pytest.raises(StateObsoleteError):
            state.run()

        with pytest.raises(StateObsoleteError):
            await state.finish()

        with pytest.raises(StateObsoleteError):
            state.reset()

        with pytest.raises(StateObsoleteError):
            state.close()

    def test_run(self, state: State):
        with pytest.raises(StateMethodError):
            state.run()

    @pytest.mark.asyncio
    async def test_finish(self, state: State):
        with pytest.raises(StateMethodError):
            await state.finish()

    def test_reset(self, state: State):
        with pytest.raises(StateMethodError):
            state.reset()

    def test_send_pdb_command(self, state: State):
        trace_id = TraceNo(1)
        command = "next"
        with pytest.raises(StateMethodError):
            state.send_pdb_command(trace_id, command)

    def test_exception(self, state: State):
        with pytest.raises(StateMethodError):
            state.exception()

    def test_result(self, state: State):
        with pytest.raises(StateMethodError):
            state.result()



from __future__ import annotations

import pytest

from nextline.state import State, Initialized, Running, Closed

from .base import BaseTestState


class TestInitialized(BaseTestState):

    state_class = Initialized

    @pytest.fixture()
    def state(self, initialized: State) -> State:
        return initialized

    @pytest.mark.asyncio
    async def test_run(self, state: State):
        running = state.run()
        assert isinstance(running, Running)
        await self.assert_obsolete(state)
        finished = await running.finish()
        finished.close()

    @pytest.mark.asyncio
    async def test_reset(self, state: State):
        reset = state.reset()
        assert isinstance(reset, Initialized)
        assert reset is not state
        await self.assert_obsolete(state)

    @pytest.mark.asyncio
    async def test_close(self, state: State):
        closed = state.close()
        assert isinstance(closed, Closed)
        await self.assert_obsolete(state)



"""Test the class Machine

TODO: add a test for the method send_pdb_command()

"""
import asyncio
import pytest

from nextline.state import Machine
from nextline.utils import SubscribableDict


SOURCE = """
import time
time.sleep(0.001)
""".strip()

SOURCE_TWO = """
x = 2
""".strip()


@pytest.mark.asyncio
async def test_init():
    obj = Machine(SOURCE)
    assert "initialized" == obj.state_name
    assert isinstance(obj.registry, SubscribableDict)
    assert SOURCE == obj.registry.get("statement")
    assert "<string>" == obj.registry.get("script_file_name")


@pytest.mark.asyncio
async def test_repr():
    obj = Machine(SOURCE)
    repr(obj)


@pytest.mark.asyncio
async def test_state_name_unknown():
    obj = Machine(SOURCE)
    obj._state = None
    assert "unknown" == obj.state_name
    del obj._state
    assert "unknown" == obj.state_name


@pytest.mark.asyncio
async def test_transitions():
    async def subscribe():
        return [y async for y in registry.subscribe("state_name")]

    obj = Machine(SOURCE)
    registry = obj.registry
    t = asyncio.create_task(subscribe())
    await asyncio.sleep(0)
    assert "initialized" == obj.state_name
    obj.run()
    assert "running" == obj.state_name
    await obj.finish()
    assert "finished" == obj.state_name
    obj.result()
    obj.exception()
    obj.reset()
    assert "initialized" == obj.state_name
    await obj.close()
    assert "closed" == obj.state_name
    expected = [
        "initialized",
        "running",
        "finished",
        "initialized",
        "closed",
    ]
    assert expected == await t


@pytest.mark.asyncio
async def test_reset_with_statement():
    async def subscribe():
        return [y async for y in registry.subscribe("state_name")]

    obj = Machine(SOURCE)
    registry = obj.registry
    t = asyncio.create_task(subscribe())
    await asyncio.sleep(0)
    assert SOURCE == obj.registry.get("statement")
    obj.run()
    await obj.finish()
    obj.reset(statement=SOURCE_TWO)
    assert SOURCE_TWO == obj.registry.get("statement")
    obj.run()
    await obj.finish()
    await obj.close()
    expected = [
        "initialized",
        "running",
        "finished",
        "initialized",
        "running",
        "finished",
        "closed",
    ]
    assert expected == await t


@pytest.fixture(autouse=True)
def monkey_patch_run(monkey_patch_run):
    yield monkey_patch_run



from __future__ import annotations

import enum
import threading
from queue import Queue
from janus import Queue as Janus, SyncQueue

from typing import (
    AsyncIterator,
    Generic,
    Literal,
    Optional,
    List,
    Tuple,
    TypeVar,
)


class _M(enum.Enum):
    # TODO: Using Enum as sentinel for now as suggested in
    # https://stackoverflow.com/a/60605919/7309855. It still has a problem. For
    # example, the type of yielded values in subscribe() is not correctly
    # inferred as _T.

    START = object()
    END = object()


_T = TypeVar("_T")


class SubscribableQueue(Generic[_T]):
    """Distribute items to subscribers

    Items can be sent from any threads. Asynchronous subscriptions don't need
    to be all in the same thread.

    A new subscriber immediately receives the latest item and then wait for new
    items.

    The order of the items is preserved.
    """

    def __init__(self):

        self._q_in: Queue[_T | Literal[_M.END]] = Queue()

        self._qs_out: List[
            Queue[Tuple[int, _T | Literal[_M.END]]]
            | SyncQueue[Tuple[int, _T | Literal[_M.END]]]
        ] = []
        self._last_enumerated: Tuple[
            int, _T | Literal[_M.START] | Literal[_M.END]
        ] = (-1, _M.START)

        self._last_item: _T | Literal[_M.START] = _M.START

        self._closed: bool = False
        self._lock_close = threading.Condition()

        self._thread = threading.Thread(target=self._listen, daemon=True)
        self._thread.start()

    @property
    def nsubscriptions(self) -> int:
        """The number of the subscribers"""
        return len(self._qs_out)

    def put(self, item: _T) -> None:
        """Send data to subscribers

        This method can be called in any thread.
        """
        if self._closed:
            raise RuntimeError(f"{self} is closed.")
        self._last_item = item
        self._q_in.put(item)

    def get(self) -> _T:
        """Most recent data that have been put"""
        if self._last_item is _M.START:
            raise LookupError
        return self._last_item

    async def subscribe(
        self,
        last: Optional[bool] = True,
    ) -> AsyncIterator[_T]:
        """Yield data as they are put

        If `last` is true, yield immediately the most recent data before
        waiting for new data.
        """
        q: Janus[Tuple[int, _T | Literal[_M.END]]] = Janus()

        self._qs_out.append(q.sync_q)

        last_idx, last_item = self._last_enumerated

        if last_item is _M.END:
            return

        try:
            if last and last_item is not _M.START:
                yield last_item

            while True:
                idx, item = await q.async_q.get()
                if item is _M.END:
                    break
                if last_idx < idx:
                    yield item

        finally:
            self._qs_out.remove(q.sync_q)

            q.close()
            await q.wait_closed()

    def close(self) -> None:
        """End gracefully"""
        with self._lock_close:
            if self._closed:
                return
            self._closed = True
            self._q_in.put(_M.END)
            self._thread.join()
            self._q_in.join()

    def _listen(self) -> None:
        """Distribution of data to subscribers

        This method runs in a thread.
        """
        idx, item = self._last_enumerated
        while item is not _M.END:
            idx += 1
            item = self._q_in.get()
            self._q_in.task_done()
            self._last_enumerated = (idx, item)
            for q in list(self._qs_out):  # list in case it changes in a thread
                q.put(self._last_enumerated)
